/**
 * This file is part of Princess, a theorem prover for Presburger
 * arithmetic with uninterpreted predicates.
 * <http://www.philipp.ruemmer.org/princess.shtml>
 *
 * Copyright (C) 2009 Philipp Ruemmer <ph_r@gmx.net>
 *
 * Princess is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Princess is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Princess.  If not, see <http://www.gnu.org/licenses/>.
 */

package ap.proof.tree;

import ap.terfor.{Formula, TermOrder}
import ap.terfor.conjunctions.Conjunction
import ap.util.{Debug, Logic}

object AndTree {
  
  private val AC = Debug.AC_PROOF_TREE

  def apply(left : ProofTree, right : ProofTree,
            vocabulary : Vocabulary,
            simplifier : ConstraintSimplifier) : ProofTree = {
    val heightDiff = heightOf(right) - heightOf(left)
    if (heightDiff > 1) {
      // then right is an <code>AndTree</code> that is very deep
      val AndTree(rightLeft, rightRight) = right
      if (heightOf(rightLeft) < heightOf(rightRight)) {
        // left-rotate
        AndTree(AndTree(left, rightLeft, vocabulary, simplifier),
                rightRight,
                vocabulary, simplifier)
      } else {
        // right-left-rotate
        val AndTree(rightLeftLeft, rightLeftRight) = rightLeft
        AndTree(AndTree(left, rightLeftLeft, vocabulary, simplifier),
                AndTree(rightLeftRight, rightRight, vocabulary, simplifier),
                vocabulary, simplifier)
      }
    } else if (heightDiff < -1) {
      // then left is an <code>AndTree</code> that is very deep
      val AndTree(leftLeft, leftRight) = left
      if (heightOf(leftRight) < heightOf(leftLeft)) {
        // right-rotate
        AndTree(leftLeft,
                AndTree(leftRight, right, vocabulary, simplifier),
                vocabulary, simplifier)
      } else {
        // left-right-rotate
        val AndTree(leftRightLeft, leftRightRight) = leftRight
        AndTree(AndTree(leftLeft, leftRightLeft, vocabulary, simplifier),
                AndTree(leftRightRight, right, vocabulary, simplifier),
                vocabulary, simplifier)
      }
    } else {
      new AndTree(left, right, vocabulary, simplifier)
    }
  }

  def apply(subtrees : Seq[ProofTree], vocabulary : Vocabulary,
            simplifier : ConstraintSimplifier) : ProofTree = {
    ////////////////////////////////////////////////////////////////////////////
    Debug.assertPre(AC, !subtrees.isEmpty)
    ////////////////////////////////////////////////////////////////////////////
    subtrees match {
      case Seq(subtree) => subtree
      case Seq(left, right) => AndTree(left, right, vocabulary, simplifier)
      case _ => {
        // divide the subtrees in two parts and recurse
        val mid = (subtrees.size + 1) / 2
        AndTree(AndTree(subtrees.slice(0, mid), vocabulary, simplifier),
                AndTree(subtrees.slice(mid, subtrees.size), vocabulary, simplifier),
                vocabulary, simplifier)
      }
    }
  }
    
  def unapply(t : ProofTree) : Option[(ProofTree, ProofTree)] = t match {
    case t : AndTree => Some((t.left, t.right))
    case _ => None
  }
  
  private def heightOf(t : ProofTree) : Int = t match {
    case t : AndTree => t.height
    case _ => 0
  }
}

class AndTree private (val left : ProofTree, val right : ProofTree,
                       val vocabulary : Vocabulary,
                       simplifier : ConstraintSimplifier) extends ProofTree {

  import AndTree.heightOf
  
  //////////////////////////////////////////////////////////////////////////////
  // Branching points in proof trees are represented as balanced binary trees
  Debug.assertCtor(AndTree.AC, (heightOf(left) - heightOf(right)).abs <= 1)
  //////////////////////////////////////////////////////////////////////////////
  
  lazy val height : Int = 1 + (heightOf(left) max heightOf(right))

  lazy val subtrees : Seq[ProofTree] = Array(left, right)
    
  lazy val closingConstraint : Conjunction =
    simplifier(Conjunction.conj(for (t <- subtrees.elements) yield t.closingConstraint,
                                order),
               order)

  lazy val closingConstantFreedom : ConstantFreedom =
    left.closingConstantFreedom meet right.closingConstantFreedom

  lazy val fixedConstantFreedom : Boolean =
    left.fixedConstantFreedom &&
    right.fixedConstantFreedom &&
    constantFreedom == closingConstantFreedom 

  lazy val stepPossible : Boolean =
    Logic.exists(for (subtree <- subtrees.elements) yield subtree.stepPossible)

  lazy val stepMeaningful : Boolean =
    (left.stepMeaningful, right.stepMeaningful) match {
    case (true, true) => true
    case (true, false) => !right.closingConstraint.isFalse
    case (false, true) => !left.closingConstraint.isFalse
    case (false, false) => false
    }

  def update(newSubtrees : Seq[ProofTree]) : ProofTree =
    AndTree(newSubtrees, vocabulary, simplifier)

  def update(newLeft : ProofTree, newRight : ProofTree,
             newConstantFreedom : ConstantFreedom) : ProofTree =
    if (left == newLeft && right == newRight &&
        constantFreedom == newConstantFreedom)
      this
    else
      AndTree(newLeft, newRight,
              vocabulary updateConstantFreedom newConstantFreedom,
              simplifier)
    
  /**
   * Given a new constant freedom for this proof tree, derive the corresponding
   * freedoms for the direct subtrees.
   */
  def newConstantFreedomForSubtrees(cf : ConstantFreedom) = (cf, cf)

  //////////////////////////////////////////////////////////////////////////////

  private def indent(str : String, prefix : String) =
    prefix + str.replaceAll("\n", "\n" + prefix)
    
  override def toString : String =
    "^ " + closingConstraint + " (/\\)\n" +
      (for (t <- subtrees) yield indent("" + t, "  ")).mkString("\n/\\\n")

}
